#include "stdafx.h"
#include "Parser.h"
#include "Units.h"
#include "ErrnoException.h"

void ShowInfo(int count, char* diskNames[]);

int main(int argc, char* argv[])
{
    ShowInfo(argc - 1, argv + 1);
    return 0;
}

const char* FsTypeToString(FsType type)
{
    switch (type)
    {
    case FsRawOrFat16:
        return "RAW or FAT16";
    case FsFat32:
        return "FAT32";
    case FsNtfs:
        return "NTFS";
    default:
        return "Unknown";
    }
}

const char* VolumeTypeToString(VolumeType type)
{
    switch (type)
    {
    case VtSimple:
        return "Simple";
    case VtSpan:
        return "Spanned";
    case VtStripe:
        return "Striped";
    case VtMirror:
        return "Mirrored";
    case VtRaid5:
        return "Raid-5";
    default:
        return "Unknown";
    }
}

uint32_t CalculatePartitionsCount(const Volume& volume)
{
    uint32_t result = 0;

    ComponentContainer::const_iterator it = volume.components.begin();
    ComponentContainer::const_iterator end = volume.components.end();

    for (; it != end; ++it)
    {
        result += (*it)->partitions.size();
    }

    return result;
}

void ShowPartitions(const Volume& volume, const DiskContainer& disks)
{
    uint32_t number = 1;
    ComponentContainer::const_iterator compIt = volume.components.begin();
    ComponentContainer::const_iterator compEnd = volume.components.end();

    for (; compIt != compEnd; ++compIt)
    {
        Component& component = **compIt;
        PartitionContainer::const_iterator partIt = component.partitions.begin();
        PartitionContainer::const_iterator partEnd = component.partitions.end();

        for (; partIt != partEnd; ++partIt, ++number)
        {
            Partition& partition = **partIt;
            DiskContainer::const_iterator disk = disks.find(partition.diskId);
            assert(disk != disks.end());

            std::cout << "\tPartition #" << number << " from disk " << disk->second.path << std::endl;
            std::cout << "\tPartition offset           : " << Units(partition.diskOffsetLba * disk->second.sectorSize) << std::endl;
            std::cout << "\tPartition size             : " << Units(partition.sizeLba * disk->second.sectorSize) << std::endl;
            if (volume.type == VtSpan || volume.type == VtSimple)
            {
                std::cout << "\tPartition offset in volume : " << Units(partition.volumeOffsetLba * disk->second.sectorSize) << std::endl;
            }
            
            std::cout << std::endl;
        }
    }
}

void ShowVolumes(const VolumeContainer& volumes, const DiskContainer& disks)
{
    std::cout << "Number of volumes: " << volumes.size() << std::endl;

    VolumeContainer::const_iterator it = volumes.begin();
    VolumeContainer::const_iterator end = volumes.end();

    for (size_t i = 1; it != end; ++it, ++i)
    {
        const Volume& volume = it->second;

        std::cout << "Volume #" << i << std::endl;
        std::cout << "Volume size          : " << Units(volume.sizeLba * 512) << std::endl;
        std::cout << "Volume file system   : " << FsTypeToString(volume.fileSystem) << std::endl;
        std::cout << "Volume type          : " << VolumeTypeToString(volume.type) << std::endl;
        std::cout << "Number of partitions : " << CalculatePartitionsCount(volume) << std::endl;

        ShowPartitions(volume, disks);
        std::cout << std::endl;
    }
}

void ShowInfo(int count, char* diskNames[])
{
    try
    {
        Parser parser;
        for (int i = 0; i < count; ++i)
        {
            parser.Parse(diskNames[i]);
        }

        parser.BuildRelations();

        const VolumeContainer& volumes = parser.GetVolumes();
        const DiskContainer& disks = parser.GetDisks();

        ShowVolumes(volumes, disks);
    }
    catch (const ErrnoException& ex)
    {
        std::cout << ex.what() << ": " << ex.GetDescription() << std::endl;
    }
    catch (const std::exception& ex)
    {
        std::cout << ex.what() << std::endl;
    }
}
